import sys
import typer

from typing import Tuple
from chris.cube.pipeline import Pipeline
from chris.cube.plugin_instance import PluginInstance


def run_pipeline_with_progress(chris_pipeline: Pipeline, parent: PluginInstance) -> Tuple[PluginInstance, ...]:
    """
    Helper to execute a pipeline with a progress bar.
    """
    root = chris_pipeline.get_root()
    with typer.progressbar(root.run(parent.id),
                           length=len(root), label='Scheduling pipeline',
                           file=sys.stderr) as proto_pipeline:
        return tuple(parent for parent, _ in proto_pipeline)



responses = {
    'https://example.com/api/v1/something/': {
        'count': 5,
        'next': 'https://example.com/api/v1/something/?limit=3&offset=3',
        'previous': None,
        'results': [
            {'id': 1}, {'id': 2}, {'id': 3}
        ],
        'collection_links': {}
    },
    'https://example.com/api/v1/something/?limit=3&offset=3': {
        'count': 5,
        'next': 'https://example.com/api/v1/something/?limit=3&offset=6',
        'previous': 'https://example.com/api/v1/something/?limit=3',
        'results': [
            {'id': 4}, {'id': 5}, {'id': 6}
        ],
        'collection_links': {}
    },
    'https://example.com/api/v1/something/?limit=3&offset=6': {
        'count': 5,
        'next': None,
        'previous': 'https://example.com/api/v1/something/?limit=3',
        'results': [
            {'id': 7}, {'id': 8}
        ],
        'collection_links': {}
    },
}



import json
from functools import cached_property
from dataclasses import dataclass, field
from typing import Generator, Collection, Dict, Tuple, Optional
from collections import deque
from chris.types import ParameterType, PluginInstanceId, PipingId
from chris.cube.plugin import Plugin
from chris.cube.plugin_instance import PluginInstance
from chris.cube.resource import ConnectedResource
from chris.cube.piping import Piping


@dataclass(frozen=True)
class PluginTree(ConnectedResource, Collection['PluginTree']):
    """
    A ``PluginTree`` is an immutable node of a directed acyclic graph
    of plugins and default parameters for each plugin.
    It usually represents a piping, its associated plugin, and
    default parameters of a runnable *ChRIS* pipeline.

    CONSTRAINT: all plugins must be associated with the same CUBE.
    """

    piping: Piping
    default_parameters: Dict[str, ParameterType]
    children: Tuple['PluginTree', ...] = field(default_factory=tuple)
    # tuple instead of frozenset because PluginTree.default_parameters
    # is a dict, which is not hashable

    def get_plugin(self) -> Plugin:
        res = self.s.get(self.piping.plugin)
        res.raise_for_status()
        return Plugin(s=self.s, **res.json())

    def run(self, plugin_instance_id: PluginInstanceId
            ) -> Generator[Tuple[PluginInstance, 'PluginTree'], None, None]:
        """
        Create plugin instances in DFS-order.
        The returned iterator must be iterated through to
        schedule this entire plugin tree.
        It produces 2-tuples of the created ``PluginInstance``
        and the ``PluginTree`` from which the instance was created.

        :param plugin_instance_id: parent plugin instance
        """
        params = {
            'previous_id': plugin_instance_id
        }
        params.update(self.default_parameters)
        created_instance = self.get_plugin().create_instance(params)
        yield created_instance, self
        for child in self.children:
            yield from child.run(created_instance.id)

    # the two traversal methods below are not currently used

    def dfs(self) -> Generator['PluginTree', None, None]:
        """
        Depth-first graph traversal.
        """
        yield self
        yield from self.children

    def bfs(self) -> Generator['PluginTree', None, None]:
        """
        Breadth-first graph traversal.

        BFS is insignificantly better than DFS because sibling
        plugin instances can be scheduled insignificantly sooner.
        A stupidly optimal solution would schedule branches by
        doing the HTTP POST requests in parallel.
        """
        queue: deque['PluginTree'] = deque()
        queue.append(self)
        while queue:
            current = queue.popleft()
            yield current
            queue.extend(current.children)

    def __iter__(self):
        return self.dfs()

    def __len__(self):
        return self.length

    @cached_property
    def length(self) -> int:
        count = 0
        for _ in self:
            count += 1
        return count

    def __contains__(self, __x: object) -> bool:
        return any(__x == e for e in self)

    def deserialize(self) -> str:
        """
        Produce this Plugin Tree as JSON according to the ``plugin_tree`` schema
        of the *ChRIS* pipeline spec.
        """
        return json.dumps(self.deserialize_tree())

    def deserialize_tree(self) -> list:
        data = []
        index_map: dict[Optional[PipingId], Optional[int]] = {
            None: None
        }

        for node in self.bfs():
            index_map[node.piping.id] = len(data)
            previous_index = index_map[node.piping.previous_id]
            data.append(node.deserialize_node(previous_index))

        return data

    def deserialize_node(self, previous_index: Optional[int]) -> dict:
        """
        Deserialize just this ``PluginTree`` (and not its children).
        """
        plugin = self.get_plugin()
        data = {
            'plugin_name': plugin.name,
            'plugin_version': plugin.version,
            'previous_index': previous_index
        }

        if self.default_parameters:
            data['plugin_parameter_defaults'] = [
                {
                    'name': name,
                    'default': value
                }
                for name, value in self.default_parameters.items()
            ]

        return data



import typer

from caw.commands.store import app, build_client


@app.command()
def export(name: str = typer.Argument(..., help='Name of pipeline.')):
    """
    Deserialize a pipeline to JSON.
    """
    client = build_client()
    pipeline = client.get_pipeline(name)
    typer.echo(pipeline.deserialize())



from dataclasses import dataclass
from typing import Optional
from chris.cube.resource import CUBEResource

from chris.types import (
    PluginUrl, CUBEUrl, PipelineId, CUBEUsername, ISOFormatDateString,
    ParameterName, ParameterType, ParameterTypeName, PipingId,
    PipelineParameterId, PluginParameterId,
    PluginName, PluginVersion, PluginId, PipingUrl
)


@dataclass(frozen=True)
class PipingParameter(CUBEResource):
    id: PipelineParameterId
    value: ParameterType
    type: ParameterTypeName
    plugin_piping_id: PipingId
    previous_plugin_piping_id: None
    param_name: ParameterName
    param_id: PluginParameterId
    plugin_piping: CUBEUrl
    plugin_name: PluginName
    plugin_version: PluginVersion
    plugin_id: PluginId
    plugin_param: CUBEUrl


@dataclass(frozen=True)
class Piping(CUBEResource):
    id: PipingId
    plugin_id: PluginId
    pipeline_id: PipelineId
    previous: Optional[PipingUrl]
    plugin: PluginUrl
    pipeline: CUBEUrl
    previous_id: Optional[PipingId] = None



import datetime
import logging
import os
import sys
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import List, Tuple, Union

import requests
import typer

from chris.client import ChrisClient
from chris.cube.files import DownloadableFile
from chris.cube.pagination import TooMuchPaginationException, MAX_REQUESTS, REQUESTS_ENV_VAR_NAME
from chris.types import CUBEUrl


def upload(client: ChrisClient, files: List[Path], parent_folder='', upload_threads=4):

    username = client.get_username()

    if parent_folder:
        upload_folder = f'{username}/uploads/{parent_folder}/{datetime.datetime.now().isoformat()}/'
    else:
        upload_folder = f'{username}/uploads/{datetime.datetime.now().isoformat()}/'

    input_files: List[Path] = []
    for path in files:
        if path.is_file():
            input_files.append(path)
        elif path.is_dir():
            nested_files = [f for f in path.rglob('**/*') if f.is_file()]
            if len(nested_files) > 0:
                input_files.extend(nested_files)
            else:
                typer.secho(f'WARNING: input directory is empty: {path}', dim=True, err=True)
        else:
            typer.secho(f'No such file or directory: {path}', fg=typer.colors.RED, err=True)
            raise typer.Abort()

    if len(input_files) == 0:
        typer.secho(f'No input files specified.', fg=typer.colors.RED, err=True)
        raise typer.Abort()

    with typer.progressbar(label='Uploading files', length=len(input_files), file=sys.stderr) as bar:
        def upload_file(input_file: str):
            client.upload(Path(input_file), Path(upload_folder))
            bar.update(1)

        with ThreadPoolExecutor(max_workers=upload_threads) as pool:
            uploads = pool.map(upload_file, input_files)

    # check for upload errors
    for upload_result in uploads:
        logging.debug(upload_result)

    typer.secho(f'Successfully uploaded {len(input_files)} files to "{upload_folder}"', fg=typer.colors.GREEN, err=True)
    return upload_folder


def download(client: ChrisClient, url: Union[str, CUBEUrl], destination: Path, threads: 4):
    """
    Download all the files from a given ChRIS API url.
    :param client: ChRIS client
    :param url: any ChRIS file resource url, e.g.
                https://cube.chrisproject.org/api/v1/uploadedfiles/
                https://cube.chrisproject.org/api/v1/uploadedfiles/?fname=chris/uploads/today
                https://cube.chrisproject.org/api/v1/3/files/
    :param destination: folder on host where to download to
    :param threads: max number of concurrent downloads
    """
    if destination.is_file():
        typer.secho(f'Cannot download into {destination}: is a file', fg=typer.colors.RED, err=True)
        raise typer.Abort()

    def __calculate_target(remote_file: DownloadableFile) -> Tuple[Path, DownloadableFile]:
        """
        Decide on a download location for a file resource in ChRIS.
        Create the parent directory if needed.
        :param remote_file: file information from ChRIS
        :return: download location on host and that file
        """
        fname = remote_file.fname
        if fname.startswith('chris/'):
            fname = fname[6:]
        target = destination.joinpath(fname)
        os.makedirs(target.parent, exist_ok=True)
        return target, remote_file

    files_to_download = _discover_files_to_download(client, url)

    with typer.progressbar(files_to_download, length=len(files_to_download), label='Getting information', file=sys.stderr) as progress:
        to_download = frozenset(__calculate_target(remote_file) for remote_file in progress)

    with typer.progressbar(length=len(to_download), label='Downloading files', file=sys.stderr) as progress:
        with ThreadPoolExecutor(max_workers=threads) as pool:

            def download_file(t: Tuple[Path, DownloadableFile], attempt=0) -> int:
                """
                Download file and move the progress bar. Retries on failure, shuts down the thread pool
                if giving up.

                :return: downloaded file size
                """
                target, remote_file = t

                if attempt >= 3:
                    typer.secho(f'Failed 3 attempts to download {remote_file.file_resource}. '
                                f'Giving up...', fg=typer.colors.RED, err=True)
                    pool.shutdown(cancel_futures=True)  # fail fast
                    raise typer.Abort()

                try:
                    remote_file.download(target)
                except requests.exceptions.RequestException as e:
                    typer.secho(f'attempt={attempt} ::: '
                                f'failed to download {remote_file.file_resource}: {str(e)}',
                                fg=typer.colors.RED, err=True)
                    return download_file(t, attempt + 1)

                if attempt > 0:
                    typer.secho(f'Successfully retried: {remote_file.file_resource}',
                                color=typer.colors.GREEN, err=True)

                progress.update(1)
                return target.stat().st_size

            sizes = pool.map(download_file, to_download)

    total_size = sum(sizes)
    if total_size < 2e5:
        size = f'{total_size} bytes'
    elif total_size < 2e8:
        size = f'{total_size / 1e6:.4f} MB'
    else:
        size = f'{total_size / 1e9:.4f} GB'
    typer.secho(size, fg=typer.colors.GREEN, err=True)


def _discover_files_to_download(client: ChrisClient, url: CUBEUrl) -> Tuple[DownloadableFile, ...]:
    typer.echo('Discovering files... ', nl=False)
    total = 0

    def report_discovered_file(f: DownloadableFile) -> DownloadableFile:
        nonlocal total
        total += 1
        typer.echo(f'\rDiscovering files... {total}', nl=False)
        return f

    try:
        search = tuple(report_discovered_file(f) for f in client.get_files(url))
    except TooMuchPaginationException:
        typer.echo(
            f'Number of paginated requests exceeded {MAX_REQUESTS}.'
            f"If you're trying to download many files, you can increase the limit:"
            f'\n\n\t {" ".join([ "env", f"{REQUESTS_ENV_VAR_NAME}={MAX_REQUESTS + 99900}"] + sys.argv)}\n'
        )
        raise typer.Abort()
    typer.echo(f'\rFound {total} files to download.')
    return search



import requests
from pytest_mock import MockerFixture
from unittest.mock import Mock
from typing import Dict


def mock_session(mocker: MockerFixture, responses: Dict[str, any]) -> Mock:
    def create_dumb_paginated_response(url: str) -> Mock:
        res = mocker.Mock()
        res.json = mocker.Mock(return_value=responses[url])
        return res

    m = mocker.Mock(spec=requests.Session)
    m.get = mocker.Mock(wraps=create_dumb_paginated_response)
    return m



import json
import abc
from dataclasses import dataclass
from chris.cube.plugin_tree import PluginTree


@dataclass(frozen=True)
class Pipeline(abc.ABC):
    name: str
    authors: str
    description: str
    category: str
    locked: bool

    @abc.abstractmethod
    def get_root(self) -> PluginTree:
        ...

    def deserialize(self) -> str:
        """
        Produce a JSON representation which can be uploaded to a CUBE instance.
        """
        data = [
            {
                'name': 'name',
                'value': self.name
            },
            {
                'name': 'authors',
                'value': self.authors,
            },
            {
                'name': 'category',
                'value': self.category,
            },
            {
                'name': 'description',
                'value': self.description,
            },
            {
                'name': 'locked',
                'value': self.locked,
            },
            {
                'name': 'plugin_tree',
                'value': self.get_root().deserialize()
            }
        ]
        template = {
            'template': {
                'data': data
            }
        }
        return json.dumps(template)



data = {
    "cancelled_jobs": 0,
    "comments": "https://example.com/api/v1/3/comments/",
    "created_jobs": 0,
    "creation_date": "2021-07-09T00:02:20.005324-04:00",
    "creator_username": "mock_username",
    "errored_jobs": 0,
    "files": "https://example.com/api/v1/3/files/",
    "finished_jobs": 4,
    "id": 3,
    "modification_date": "2021-07-09T00:02:20.005357-04:00",
    "name": "A mock feed",
    "note": "https://example.com/api/v1/note3/",
    "owner": [
        "https://example.com/api/v1/users/4/"
    ],
    "plugin_instances": "https://example.com/api/v1/3/plugininstances/",
    "registering_jobs": 0,
    "scheduled_jobs": 0,
    "started_jobs": 0,
    "taggings": "https://example.com/api/v1/3/taggings/",
    "tags": "https://example.com/api/v1/3/tags/",
    "template": {
        "data": [
            {
                "name": "name",
                "value": ""
            },
            {
                "name": "owner",
                "value": ""
            }
        ]
    },
    "url": "https://example.com/api/v1/3/",
    "waiting_jobs": 0
}



# TODO got lazy...

data = {
    'https://example.com/api/v1/plugins/4/': {},

    'https://example.com/api/v1/plugins/5/': {},

    'https://example.com/api/v1/plugins/6/': {},

    'https://example.com/api/v1/pipelines/': {
        "collection_links": {
            "plugins": "https://example.com/api/v1/plugins/"
        },
        "count": 2,
        "next": None,
        "previous": None,
        "queries": [
            {
                "data": [
                    {
                        "name": "id",
                        "value": ""
                    },
                    {
                        "name": "owner_username",
                        "value": ""
                    },
                    {
                        "name": "name",
                        "value": ""
                    },
                    {
                        "name": "category",
                        "value": ""
                    },
                    {
                        "name": "description",
                        "value": ""
                    },
                    {
                        "name": "authors",
                        "value": ""
                    },
                    {
                        "name": "min_creation_date",
                        "value": ""
                    },
                    {
                        "name": "max_creation_date",
                        "value": ""
                    }
                ],
                "href": "https://example.com/api/v1/pipelines/search/",
                "rel": "search"
            }
        ],
        "results": [
            {
                "authors": "Jennings Zhang <Jennings.Zhang@childrens.harvard.edu>",
                "category": "",
                "creation_date": "2021-11-12T23:03:58.401890-05:00",
                "default_parameters": "https://example.com/api/v1/pipelines/1/parameters/",
                "description": "A linear and boring pipeline.",
                "id": 1,
                "instances": "https://example.com/api/v1/pipelines/1/instances/",
                "locked": False,
                "modification_date": "2021-11-12T23:03:58.401912-05:00",
                "name": "Example linear pipeline",
                "owner_username": "chris",
                "plugin_pipings": "https://example.com/api/v1/pipelines/1/pipings/",
                "plugins": "https://example.com/api/v1/pipelines/1/plugins/",
                "url": "https://example.com/api/v1/pipelines/1/"
            },
            {
                "authors": "Jennings Zhang <Jennings.Zhang@childrens.harvard.edu>",
                "category": "",
                "creation_date": "2021-11-12T23:03:58.742417-05:00",
                "default_parameters": "https://example.com/api/v1/pipelines/2/parameters/",
                "description": "A fun but nonetheless useless pipeline.",
                "id": 2,
                "instances": "https://example.com/api/v1/pipelines/2/instances/",
                "locked": False,
                "modification_date": "2021-11-12T23:03:58.742439-05:00",
                "name": "Example branching pipeline",
                "owner_username": "chris",
                "plugin_pipings": "https://example.com/api/v1/pipelines/2/pipings/",
                "plugins": "https://example.com/api/v1/pipelines/2/plugins/",
                "url": "https://example.com/api/v1/pipelines/2/"
            }
        ],
        "template": {
            "data": [
                {
                    "name": "name",
                    "value": ""
                },
                {
                    "name": "authors",
                    "value": ""
                },
                {
                    "name": "category",
                    "value": ""
                },
                {
                    "name": "description",
                    "value": ""
                },
                {
                    "name": "locked",
                    "value": ""
                },
                {
                    "name": "plugin_tree",
                    "value": ""
                },
                {
                    "name": "plugin_inst_id",
                    "value": ""
                }
            ]
        }
    },

    'https://example.com/api/v1/pipelines/2/': {
        "authors": "Jennings Zhang <Jennings.Zhang@childrens.harvard.edu>",
        "category": "",
        "creation_date": "2021-11-12T23:03:58.742417-05:00",
        "default_parameters": "https://example.com/api/v1/pipelines/2/parameters/",
        "description": "A fun but nonetheless useless pipeline.",
        "id": 2,
        "instances": "https://example.com/api/v1/pipelines/2/instances/",
        "locked": False,
        "modification_date": "2021-11-12T23:03:58.742439-05:00",
        "name": "Example branching pipeline",
        "owner_username": "chris",
        "plugin_pipings": "https://example.com/api/v1/pipelines/2/pipings/",
        "plugins": "https://example.com/api/v1/pipelines/2/plugins/",
        "template": {
            "data": [
                {
                    "name": "name",
                    "value": ""
                },
                {
                    "name": "authors",
                    "value": ""
                },
                {
                    "name": "category",
                    "value": ""
                },
                {
                    "name": "description",
                    "value": ""
                }
            ]
        },
        "url": "https://example.com/api/v1/pipelines/2/"
    },

    'https://exmaple.com/api/v1/pipelines/2/pipings/': {
        "collection_links": {
            "pipeline": "https://example.com/api/v1/pipelines/2/"
        },
        "count": 8,
        "next": None,
        "previous": None,
        "results": [
            {
                "id": 4,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": None,
                "url": "https://example.com/api/v1/pipelines/pipings/4/"
            },
            {
                "id": 5,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": "https://example.com/api/v1/pipelines/pipings/4/",
                "previous_id": 4,
                "url": "https://example.com/api/v1/pipelines/pipings/5/"
            },
            {
                "id": 6,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": "https://example.com/api/v1/pipelines/pipings/4/",
                "previous_id": 4,
                "url": "https://example.com/api/v1/pipelines/pipings/6/"
            },
            {
                "id": 7,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": "https://example.com/api/v1/pipelines/pipings/5/",
                "previous_id": 5,
                "url": "https://example.com/api/v1/pipelines/pipings/7/"
            },
            {
                "id": 8,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": "https://example.com/api/v1/pipelines/pipings/6/",
                "previous_id": 6,
                "url": "https://example.com/api/v1/pipelines/pipings/8/"
            },
            {
                "id": 9,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": "https://example.com/api/v1/pipelines/pipings/6/",
                "previous_id": 6,
                "url": "https://example.com/api/v1/pipelines/pipings/9/"
            },
            {
                "id": 10,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": "https://example.com/api/v1/pipelines/pipings/6/",
                "previous_id": 6,
                "url": "https://example.com/api/v1/pipelines/pipings/10/"
            },
            {
                "id": 11,
                "pipeline": "https://example.com/api/v1/pipelines/2/",
                "pipeline_id": 2,
                "plugin": "https://example.com/api/v1/plugins/4/",
                "plugin_id": 4,
                "previous": "https://example.com/api/v1/pipelines/pipings/10/",
                "previous_id": 10,
                "url": "https://example.com/api/v1/pipelines/pipings/11/"
            }
        ]
    },

    'https://example.com/api/v1/pipelines/2/parameters/': {
        "count": 40,
        "next": None,
        "previous": None,
        "results": [
            {
                "id": 8,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/4/",
                "plugin_piping_id": 4,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": None,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/8/",
                "value": "0"
            },
            {
                "id": 7,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/4/",
                "plugin_piping_id": 4,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": None,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/7/",
                "value": "a"
            },
            {
                "id": 10,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/5/",
                "plugin_piping_id": 5,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/10/",
                "value": "0"
            },
            {
                "id": 9,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/5/",
                "plugin_piping_id": 5,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/9/",
                "value": "b"
            },
            {
                "id": 12,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/6/",
                "plugin_piping_id": 6,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/12/",
                "value": "0"
            },
            {
                "id": 11,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/6/",
                "plugin_piping_id": 6,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/11/",
                "value": "c"
            },
            {
                "id": 14,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/7/",
                "plugin_piping_id": 7,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 5,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/14/",
                "value": "0"
            },
            {
                "id": 13,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/7/",
                "plugin_piping_id": 7,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 5,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/13/",
                "value": "d"
            },
            {
                "id": 16,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/8/",
                "plugin_piping_id": 8,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/16/",
                "value": "0"
            },
            {
                "id": 15,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/8/",
                "plugin_piping_id": 8,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/15/",
                "value": "e"
            },
            {
                "id": 18,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/9/",
                "plugin_piping_id": 9,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/18/",
                "value": "0"
            },
            {
                "id": 17,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/9/",
                "plugin_piping_id": 9,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/17/",
                "value": "f"
            },
            {
                "id": 20,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/10/",
                "plugin_piping_id": 10,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/20/",
                "value": "0"
            },
            {
                "id": 19,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/10/",
                "plugin_piping_id": 10,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/19/",
                "value": "g"
            },
            {
                "id": 22,
                "param_id": 11,
                "param_name": "sleepLength",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/11/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/11/",
                "plugin_piping_id": 11,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 10,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/22/",
                "value": "0"
            },
            {
                "id": 21,
                "param_id": 9,
                "param_name": "prefix",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/9/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/11/",
                "plugin_piping_id": 11,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 10,
                "type": "string",
                "url": "https://example.com/api/v1/pipelines/string-parameter/21/",
                "value": "h"
            },
            {
                "id": 4,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/4/",
                "plugin_piping_id": 4,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": None,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/4/",
                "value": 1
            },
            {
                "id": 5,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/5/",
                "plugin_piping_id": 5,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/5/",
                "value": 1
            },
            {
                "id": 6,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/6/",
                "plugin_piping_id": 6,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/6/",
                "value": 1
            },
            {
                "id": 7,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/7/",
                "plugin_piping_id": 7,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 5,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/7/",
                "value": 1
            },
            {
                "id": 8,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/8/",
                "plugin_piping_id": 8,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/8/",
                "value": 1
            },
            {
                "id": 9,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/9/",
                "plugin_piping_id": 9,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/9/",
                "value": 1
            },
            {
                "id": 10,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/10/",
                "plugin_piping_id": 10,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/10/",
                "value": 1
            },
            {
                "id": 11,
                "param_id": 12,
                "param_name": "dummyInt",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/12/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/11/",
                "plugin_piping_id": 11,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 10,
                "type": "integer",
                "url": "https://example.com/api/v1/pipelines/integer-parameter/11/",
                "value": 1
            },
            {
                "id": 4,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/4/",
                "plugin_piping_id": 4,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": None,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/4/",
                "value": 1.1
            },
            {
                "id": 5,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/5/",
                "plugin_piping_id": 5,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/5/",
                "value": 1.1
            },
            {
                "id": 6,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/6/",
                "plugin_piping_id": 6,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/6/",
                "value": 1.1
            },
            {
                "id": 7,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/7/",
                "plugin_piping_id": 7,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 5,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/7/",
                "value": 1.1
            },
            {
                "id": 8,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/8/",
                "plugin_piping_id": 8,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/8/",
                "value": 1.1
            },
            {
                "id": 9,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/9/",
                "plugin_piping_id": 9,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/9/",
                "value": 1.1
            },
            {
                "id": 10,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/10/",
                "plugin_piping_id": 10,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/10/",
                "value": 1.1
            },
            {
                "id": 11,
                "param_id": 13,
                "param_name": "dummyFloat",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/13/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/11/",
                "plugin_piping_id": 11,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 10,
                "type": "float",
                "url": "https://example.com/api/v1/pipelines/float-parameter/11/",
                "value": 1.1
            },
            {
                "id": 4,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/4/",
                "plugin_piping_id": 4,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": None,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/4/",
                "value": False
            },
            {
                "id": 5,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/5/",
                "plugin_piping_id": 5,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/5/",
                "value": False
            },
            {
                "id": 6,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/6/",
                "plugin_piping_id": 6,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 4,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/6/",
                "value": False
            },
            {
                "id": 7,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/7/",
                "plugin_piping_id": 7,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 5,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/7/",
                "value": False
            },
            {
                "id": 8,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/8/",
                "plugin_piping_id": 8,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/8/",
                "value": False
            },
            {
                "id": 9,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/9/",
                "plugin_piping_id": 9,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/9/",
                "value": False
            },
            {
                "id": 10,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/10/",
                "plugin_piping_id": 10,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 6,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/10/",
                "value": False
            },
            {
                "id": 11,
                "param_id": 10,
                "param_name": "b_ignoreInputDir",
                "plugin_id": 4,
                "plugin_name": "pl-simpledsapp",
                "plugin_param": "https://example.com/api/v1/plugins/parameters/10/",
                "plugin_piping": "https://example.com/api/v1/pipelines/pipings/11/",
                "plugin_piping_id": 11,
                "plugin_version": "2.0.2",
                "previous_plugin_piping_id": 10,
                "type": "boolean",
                "url": "https://example.com/api/v1/pipelines/boolean-parameter/11/",
                "value": False
            }
        ]
    }
}
